﻿namespace App291
{
  using Microsoft.Graphics.Canvas.Effects;
  using System;
  using System.Numerics;
  using System.Threading.Tasks;
  using Windows.Foundation;
  using Windows.Graphics.Imaging;
  using Windows.UI.Composition;
  using Windows.UI.Xaml;
  using Windows.UI.Xaml.Controls;
  using Windows.UI.Xaml.Hosting;
  using Windows.UI.Xaml.Media.Imaging;

  public static class PanelExtensions
  {
    class EffectContext : IDisposable
    {
      // no finalizer as I have no unmanaged resources that I'm
      // aware.
      public void Dispose()
      {
        // tagging this onto disposal is a bit naughty but there was a 
        // need to dipose the other bits so this removes the SpriteVisual
        // from the ContainerVisual it's living in.
        this.spriteVisual.Parent.Children.RemoveAll();

        // these bits all need disposal...
        this.effectBrush.Dispose();
        this.imageBrush.Dispose();
        this.spriteVisual.Dispose();
      }
      public SpriteVisual spriteVisual;
      public CompositionImageBrush imageBrush;
      public CompositionEffectBrush effectBrush;
    }
    static ContainerVisual EnsureContainerVisualInXamlPanel(Panel panel)
    {
      var containerVisual =
        ElementCompositionPreview.GetElementChildVisual(panel) as ContainerVisual;

      if (containerVisual == null)
      {
        Compositor compositor = ElementCompositionPreview.GetElementVisual(panel).Compositor;

        containerVisual = compositor.CreateContainerVisual();
        ElementCompositionPreview.SetElementChildVisual(panel, containerVisual);
      }
      return (containerVisual);
    }
    static async Task<SoftwareBitmap> RenderXamlElementToBGRASoftwareBitmapAsync(UIElement element)
    {
      var renderedBitmap = new RenderTargetBitmap();

      // Unsure of the behaviour here because if I specify a width, height they seem to
      // be ignored here so leaving those parameters out and then resizing it myself
      // later on.
      await renderedBitmap.RenderAsync(element);

      // Get the pixels out of that bitmap as BGRA8
      var pixels = await renderedBitmap.GetPixelsAsync();

      // Copy into a software bitmap. 
      var softwareBitmap = SoftwareBitmap.CreateCopyFromBuffer(
        pixels,
        BitmapPixelFormat.Bgra8,
        renderedBitmap.PixelWidth,
        renderedBitmap.PixelHeight,
        BitmapAlphaMode.Premultiplied);

      return (softwareBitmap);
    }
    public static async Task<IDisposable> SaturateAsync(
      this Panel panel, int saturationPercentage)
    {
      var compositor =
        ElementCompositionPreview.GetElementVisual(panel).Compositor;

      // We put a container visual into the Xaml panel if we haven't
      // done that already. The container will stay there forever
      // because I don't know how to remove it. But...we'll add/remove
      // a child sprite visual to it.
      var containerVisual = EnsureContainerVisualInXamlPanel(panel);

      // the return value which provides a way for our caller to
      // dispose of what we give them when they are done.
      var returnContext = new EffectContext();

      // this is not necessarily the same width/height as the
      // rendered bitmap we're about to make.
      var panelSize = new Size(panel.ActualWidth, panel.ActualHeight);

      // Render the XAML Panel to a software bitmap via a
      // RenderTargetBitmap.
      using (var renderedSoftwareBitmap =
        await RenderXamlElementToBGRASoftwareBitmapAsync(panel))
      {
        // Create an instance of my 'composition image brush' from that.
        returnContext.imageBrush = CompositionImageBrush.FromBGRASoftwareBitmap(
          compositor,
          renderedSoftwareBitmap,
          panelSize);
      }

      // Create a saturation effect brush taking our image brush as
      // the source.
      returnContext.effectBrush = CreateSaturationEffectBrushWithSource(
        compositor, saturationPercentage / 100.0f, returnContext.imageBrush.Brush);

      // Create a new Sprite Visual to draw our image.
      returnContext.spriteVisual = compositor.CreateSpriteVisual();

      returnContext.spriteVisual.Size =
        new Vector2((float)panelSize.Width, (float)panelSize.Height);

      // We want to paint our visual with our effect brush
      returnContext.spriteVisual.Brush = returnContext.effectBrush;

      // and add it to the container visual that's now in the panel.
      containerVisual.Children.InsertAtTop(returnContext.spriteVisual);

      // Quite a lot of stuff now needs disposing!
      return (returnContext);
    }

    static CompositionEffectBrush CreateSaturationEffectBrushWithSource(
      Compositor compositor, 
      float saturationAmount,
      CompositionBrush sourceBrush)
    {
      CompositionEffectBrush brush = null;

      using (
        var saturationEffect = new SaturationEffect()
        {
          Source = new CompositionEffectSourceParameter("source"),
          Saturation = saturationAmount
        })
      {
        // So that we can make a factory
        using (var effectFactory = compositor.CreateEffectFactory(saturationEffect))
        {
          // SO that we can make a brush
          brush = effectFactory.CreateBrush();

          // which uses our drawing brush as its source
          brush.SetSourceParameter("source", sourceBrush);
        }
      }
      return (brush);
    }
  }
}
